from __future__ import generators
import re

class Board:
    def load(self, server, robot):
        self.server = server
        self.robot = robot
        robot.board = self
        server.write('Player\n')

        self.tokens = tokens = server.tokens()

        self.width = w = int(tokens.next())
        self.height = int(tokens.next())
        assert tokens.next() is None, 'Protocol EOL error!'

        self.move_map = {'N': w, 'S': -w, 'E': 1, 'W': -1}

        self.possibles = possibles = {}
        self.regions = regions = {}
        rows = []
        regn = 2
        stop_region = '\0\0'
        last_row = ['\0\1'] * w
        for rown in range(self.height):
            row = tokens.next()
            assert tokens.next() is None, 'Protocol EOL error!'
            i = row.find('@')
            while i >= 0:
                possibles[rown * w + i] = 1
                i = row.find('@', i + 1)
            sqs = [stop_region] * w
            runs = re.split('([#~]+)', row)
            start = 0
            while runs:
                run = len(runs.pop(0))
                if run:
                    r = last_row[start]
                    if (r == stop_region or
                        (start and r == last_row[start - 1])):
                        old_r = r
                        r = chr(regn / 256) + chr(regn % 256)
                        regn += 1
                        if old_r <> stop_region:
                            self.add_arc(old_r, r, rown * w + start, 1, w)
                    sqs[start] = r
                    old_r = r
                i = start + 1
                start += run
                while i < start:
                    r = last_row[i]
                    if r == last_row[i - 1]:
                        if r <> stop_region and r <> old_r:
                            self.add_arc(r, old_r, rown * w + i, 1, w)
                        r = old_r
                    else:
                        if r == stop_region:
                            r = chr(regn / 256) + chr(regn % 256)
                            regn += 1
                        if r <> old_r:
                            self.add_arc(old_r, r, rown * w + i, w, 1)
                    sqs[i] = r
                    i += 1
                    old_r = r
                if runs:
                    start += len(runs.pop(0))
            last_row = sqs
            rows.append(''.join(sqs))
        self.map = ''.join(rows)

        robot.id, robot.capacity, robot.money = map(
            int, [tokens.next(), tokens.next(), tokens.next()])
        assert tokens.next() is None, 'Protocol EOL error!'

        self.robots = {}
        self.held_packages = {}
        self.free_packages = {}

    def add_arc(self, region1, region2, at, h, v):
        arcs1 = self.regions.setdefault(region1, {})
        arcs2 = self.regions.setdefault(region2, {})
        if arcs1.has_key(region2):
            arcs1[region2][1] += h
            arcs2[region1][1] += h
        else:
            arcs1[region2] = [at - v, at - v]
            arcs2[region1] = [at, at]

    def update_packages(self):
        # We now see this location's actual packages
        location = self.robot.location
        if self.possibles.has_key(location):
            del self.possibles[location]
        free_packages = self.free_packages
        if free_packages.has_key(location):
            del free_packages[location]
        tokens = self.tokens
        token = tokens.next()
        if token == 'Robot':
            raise 'Death'
        w = self.width
        packages = []
        while token is not None:
            x, y, weight = [int(tokens.next()) for i in range(3)]
            yield int(token), weight, x - 1 + (y - 1) * w
            packages.append(int(token))
            token = tokens.next()
        if packages:
            free_packages[location] = packages

    def get_response(self):
        my_id = self.robot.id
        tokens = self.tokens
        move_map = self.move_map
        robots = self.robots
        unseen = robots.copy()

        token = tokens.next()
        if token == 'Robot':
            raise 'Death'
        while token is not None:
            assert token[0]=='#', 'Protocol robot id error!'
            id = int(token[1:])
            if unseen.has_key(id):
                del unseen[id]

            token = tokens.next()
            while token is not None:
                if token[0]=='#':
                    break
                elif token in 'NSEW':
                    robots[id] += move_map[token]
                elif token == 'X':
                    x = int(tokens.next()) - 1
                    assert tokens.next()=='Y', 'Protocol X/Y error!'
                    robots[id] = x + (int(tokens.next()) - 1) * self.width
                elif token == 'P':
                    p_id = int(tokens.next())
                    self.held_packages[p_id] = id
                    try:
                        addr = robots[id]
                        self.free_packages[addr].remove(p_id)
                        if not self.free_packages[addr]:
                            del self.free_packages[addr]
                    except:
                        pass
                    if id == my_id:
                        self.robot.picked_up(p_id)
                elif token == 'D':
                    p_id = int(tokens.next())
                    self.free_packages.setdefault(robots[id], []).append(p_id)
                    try:
                        del self.held_packages[p_id]
                    except:
                        pass
                    if id == my_id:
                        self.robot.dropped(p_id)
                else:
                    assert 0, 'Protocol unknown update error!'
                token = tokens.next()
        for id in unseen.keys():
            del robots[id]
        self.robot.location = self.robots[my_id]

    def region_of(self, location):
        location *= 2
        return self.map[location:location + 2]

    def reachable(self, location):
        location *= 2
        region = self.map[location:location + 2]
        if self.regions.has_key(region):
            return region

    def plan(self, location, goal):
        region1 = self.region_of(location)
        region2 = self.region_of(goal)
        if region1 == region2:
            return [region1], self.simple_cost(location, goal)

        regions = self.regions
        simple_cost = self.simple_cost
        region_step = self.region_step
        one_step = self.one_step
        v = {region1: (None, 0)}
        done = v.has_key
        edge = [(0, region1, location)]
        while edge:
            added = 0
            cost, r, at = edge.pop()
            for target, boundary in regions[r].items():
                if done(target):
                    continue
                new_at = one_step(region_step(at, target), target)
                new_cost = cost - simple_cost(at, new_at)
                v[target] = (r, new_cost)
                if target == region2:
                    edge = None
                    break
                edge.append((new_cost, target, new_at))
                added = 1
            else:
                if added:
                    edge.sort()
        r, cost = v[region2]
        plan = [region2]
        while r is not None:
            plan.insert(0, r)
            r = v[r][0]
        return plan, -cost
    
    def simple_step(self, location, goal):
        w = self.width
        xdelta = (goal % w) - (location % w)
        ydelta = (goal / w) - (location / w)
        reachable = self.reachable
        xregion = yregion = None
        if xdelta > 0:
            xregion = reachable(location + 1)
            xmove = 'E'
        elif xdelta < 0:
            xregion = reachable(location - 1)
            xmove = 'W'
        if ydelta > 0:
            yregion = reachable(location + w)
            ymove = 'N'
        elif ydelta < 0:
            yregion = reachable(location - w)
            ymove = 'S'
        if xregion is None:
            if yregion is None:
                return
            return '1 Move %s\n' % ymove
        if yregion is None:
            return '1 Move %s\n' % xmove
        region = self.region_of(location)
        if xregion == region:
            return '1 Move %s\n' % xmove
        return '1 Move %s\n' % ymove

    def one_step(self, location, region_to):
        for move in self.move_map.values():
            goal = location + move
            if goal > 0 and self.region_of(goal) == region_to:
                return goal
        return location
    
    def region_step(self, location, region_to):
        region_of = self.region_of
        region_from = region_of(location)
        tl, br = self.regions[region_from][region_to]
        if tl == br:
            if tl == location:
                return self.one_step(location, region_to)
            return tl
        w = self.width
        x = location % w
        bx = tl % w
        # Vertical boundary
        if bx == br % w:
            if location < tl:
                return tl
            if location > br:
                return br
            if x == bx:
                return self.one_step(location, region_to)
            return bx - x + location
        # Horizontal boundary
        if x < bx:
            return tl
        if x > br % w:
            return br
        if tl <= location <= br:
            return self.one_step(location, region_to)
        return x - bx + tl

    def simple_cost(self, loc1, loc2):
        w = self.width
        return abs((loc1 % w) - (loc2 % w)) + abs((loc1 / w) - (loc2 / w))

    def prune(self):
        region = self.region_of(self.robot.location)
        regions = self.regions
        v = {region: regions[region]}
        done = v.has_key
        edge = [region]
        while edge:
            r = edge.pop()
            for target in regions[r].keys():
                if done(target):
                    continue
                v[target] = regions[target]
                edge.append(target)
        self.regions = v

    def dump_map(self):
        map = self.map
        location = self.robot.location
        w = self.width
        for i in range(self.height):
            chars = []
            for j in range(w):
                addr = (i * w + j)
                if addr == location:
                    chars.append(' ')
                else:
                    chars.append( `map[addr * 2 + 1]`[-2] )
            print ''.join(chars)
