import sys, random, time

class Robot:
    def run(self):
        self.plan = None
        self.goal = None
        self.cargo = {}
        self.available = self.capacity
        self.package_info = {}
        self.base_time = time.time()
        self.commands = 0
        self.board.get_response()
        self.board.prune()
        while 1:
            self.see_packages()
            command = (
                self.drop_packages() or
                self.pick_packages() or
                self.deliver_package() or
                self.find_package()
                )
            self.board.server.write(command)
            self.commands += 1
            self.board.get_response()

    def see_packages(self):
        board = self.board
        package_info = self.package_info
        candidates = self.candidates = []
        for p_id, weight, address in board.update_packages():
            package_info[p_id] = (weight, address)
            candidates.append((p_id, weight, address))

    def drop_packages(self):
        package_info = self.package_info
        location = self.location
        drops = self.cargo.get(location)
        if drops:
            return '1 Drop %s\n' % ' '.join(map(str, drops))

    def time_to_burn(self):
        return self.commands > time.time() - self.base_time + 10

    def pick_packages(self):
        if self.time_to_burn():
            return self.fancy_pick_packages()
        picks = []
        reachable = self.board.reachable
        available = self.available
        for p_id, weight, address in self.candidates:
            if weight <= available and reachable(address):
                picks.append(str(p_id))
                available -= weight
        if picks:
            return '1 Pick %s\n' % ' '.join(picks)

    def fancy_pick_packages(self):
        packs = []
        reachable = self.board.reachable
        available = self.available
        time_to_burn = self.time_to_burn
        location = self.location
        plan = self.board.plan
        for p_id, weight, address in self.candidates:
            if weight <= available and reachable(address):
                p, cost = plan(location, address)
                score = weight / (cost + 10)
                packs.append((score, weight, p_id, address, p))
            if not time_to_burn():
                break
        packs.sort()
        picks = []
        while packs:
            score, weight, p_id, address, p = packs.pop()
            if weight <= available:
                picks.append(str(p_id))
                available -= weight
                if len(picks) == 1:
                    self.goal = address
                    self.path = p
        if picks:
            return '1 Pick %s\n' % ' '.join(picks)

    def deliver_package(self):
        goal = self.goal
        if goal is None or goal == self.location:
            if not self.cargo:
                return
            if self.time_to_burn() and len(self.cargo) > 1:
                goal = self.plan_delivery()
            else:
                goal = self.cargo.keys()[0]
            self.goal = goal
        return self.goal_move(goal)

    def plan_delivery(self):
        choices = []
        plan = self.board.plan
        location = self.location
        addresses = self.cargo.keys()
        opt_goal = addresses.pop()
        opt_p, min_cost = plan(location, opt_goal)
        for address in addresses:
            p, cost = plan(location, address)
            if cost < min_cost:
                opt_goal, opt_p, min_cost = address, p, cost
        self.plan = opt_p
        return opt_goal

    def goal_move(self, goal):
        location = self.location
        region = self.board.region_of(location)
        region2 = self.board.region_of(goal)
        plan = self.plan
        if plan is None or region not in plan or region2 not in plan:
            plan, cost = self.board.plan(location, goal)
            self.plan = plan
        if region <> plan[0]:
            del plan[:plan.index(region)]
        if len(plan) > 1:
            goal = self.board.region_step(location, plan[1])
        step = self.board.simple_step(location, goal)
        if step is None:
            self.board.simple_step(location, goal)
            self.plan = None
            return '1 Drop\n'
        return step

    def find_package(self):
        possibles = self.board.possibles
        if possibles:
            self.goal = goal = random.choice(possibles.keys())
            return self.goal_move(goal)
        free_packages = self.board.free_packages
        if free_packages:
            self.goal = goal = random.choice(free_packages.keys())
            return self.goal_move(goal)
        regions = self.board.regions
        region = random.choice(regions.keys())
        return self.goal_move(regions[region].values()[0][0])

    def picked_up(self, p_id):
        weight, address = self.package_info[p_id]
        self.cargo.setdefault(address, []).append(p_id)
        self.available -= weight

    def dropped(self, p_id):
        weight, address = self.package_info[p_id]
        packs = self.cargo[address]
        packs.remove(p_id)
        self.available += weight
        if not packs:
            del self.cargo[address]
            if address == self.goal:
                self.goal = self.plan = None
